From 7378526770b823cf76ca66d31ad6aa079dd15cec Mon Sep 17 00:00:00 2001
From: Niklas Haas <git@haasn.dev>
Date: Sat, 19 Nov 2022 00:26:46 +0100
Subject: [PATCH] shaders/colorspace: fix tone mapping desaturation

The desaturation code was seemingly completely wrong. I have no idea
what it was even doing, but it's clear that whatever the intent was, it
was not working correctly in practice.

Fixes: https://github.com/mpv-player/mpv/issues/10844
---
 src/shaders/colorspace.c | 22 +++++++++++++++-------
 1 file changed, 15 insertions(+), 7 deletions(-)

diff --git a/src/shaders/colorspace.c b/src/shaders/colorspace.c
index 0090bc6b..63e18336 100644
--- a/src/shaders/colorspace.c
+++ b/src/shaders/colorspace.c
@@ -1462,13 +1462,21 @@ static void tone_map(pl_shader sh,
             .data = PL_TRANSPOSE_3X3(rgb2xyz.m),
         }));
 
-        // Tuned to meet the desired desaturation at 1000 -> SDR
-        float desat = dst_max > src_max ? 1.075f : 1.1f;
-        float exponent = logf(desat) / logf(1000 / PL_COLOR_SDR_WHITE);
-        GLSL("float orig = max(xyz.y, %s);                      \n"
-             "xyz.y = tone_map(xyz.y);                          \n"
-             "xyz.xz *= pow(xyz.y / orig, %s) * xyz.y / orig;   \n",
-             SH_FLOAT(dst_min), SH_FLOAT(exponent));
+        // Tuned to meet the desired desaturation at 1000 <-> SDR
+        const float ref_ratio = 1000 / PL_COLOR_SDR_WHITE;
+        float ratio = src_max / dst_max, desat;
+        if (ratio >= 1) {
+            float k = (ratio - 1) / (ref_ratio - 1);
+            desat = PL_MIX(1.0f, 1/1.1f, k);
+        } else {
+            float k = (1 - ratio) / (1 - 1 / ref_ratio);
+            desat = PL_MIX(1.0f, 1.075f, k);
+        }
+
+        GLSL("float orig = max(xyz.y, %s);  \n"
+             "xyz.y = tone_map(xyz.y);      \n"
+             "xyz.xz *= %s * xyz.y / orig;  \n",
+             SH_FLOAT(dst_min), SH_FLOAT(desat));
 
         // Extra luminance correction when reducing dynamic range
         if (src_max > dst_max)
